PID_COL = "PID"
TREATMENT_COL = "treatment"
OUTCOME_COL = "outcome"
PS_COL = "ps"
PROBAS_T1_COL = "probas_t1"  # probability of outcome when treated
PROBAS_T0_COL = "probas_t0"  # probability of outcome when not treated
PROBAS_COL = "probas"  # probability of outcome under given treatment


OUTCOME_COL = "Y"
OUTCOME_CF_COL = "Y_cf"

TREATED_PID_COL = "treated_pid_col"
CONTROL_PID_COL = "control_pid_col"
DISTANCE_COL = "distance_col"
